@@ -157,6 +157,9 @@ int lookup_nss_read_master(struct master *master, time_t age)
 			result = do_read_master(master, "file", age);
 		}
 
+		if (result == NSS_STATUS_UNAVAIL)
+			master->read_fail = 1;
+
 		return !result;
 	} else {
 		char *name = master->name;
@@ -173,7 +176,9 @@ int lookup_nss_read_master(struct master *master, time_t age)
 			    !strncmp(name, "nis:", 4) ||
 			    !strncmp(name, "nisplus:", 8) ||
 			    !strncmp(name, "ldap:", 5) ||
-			    !strncmp(name, "ldaps:", 6)) {
+			    !strncmp(name, "ldaps:", 6) ||
+			    !strncmp(name, "sss:", 4) ||
+			    !strncmp(name, "dir:", 4)) {
 				strncpy(source, name, tmp - name);
 
 				/*
@@ -194,6 +199,9 @@ int lookup_nss_read_master(struct master *master, time_t age)
 				result = do_read_master(master, source, age);
 				master->name = name;
 
+				if (result == NSS_STATUS_UNAVAIL)
+					master->read_fail = 1;
+
 				return !result;
 			}
 		}
@@ -248,6 +256,9 @@ int lookup_nss_read_master(struct master *master, time_t age)
 			continue;
 		}
 
+		if (result == NSS_STATUS_UNAVAIL)
+			master->read_fail = 1;
+
 		status = check_nss_result(this, result);
 		if (status >= 0) {
 			free_sources(&nsslist);
@@ -286,14 +297,20 @@ static int do_read_map(struct autofs_point *ap, struct map_source *map, time_t a
 
 	status = lookup->lookup_read_map(ap, age, lookup->context);
 
-	map->stale = 0;
+	if (status != NSS_STATUS_SUCCESS)
+		map->stale = 0;
 
 	/*
 	 * For maps that don't support enumeration return success
 	 * and do whatever we must to have autofs function with an
 	 * empty map entry cache.
+	 *
+	 * For indirect maps that use the browse option, when the
+	 * server is unavailable continue as best we can with
+	 * whatever we have in the cache, if anything.
 	 */
-	if (status == NSS_STATUS_UNKNOWN)
+	if (status == NSS_STATUS_UNKNOWN ||
+	   (ap->type == LKP_INDIRECT && status == NSS_STATUS_UNAVAIL))
 		return NSS_STATUS_SUCCESS;
 
 	return status;
@@ -396,8 +413,10 @@ static enum nsswitch_status read_map_source(struct nss_source *this,
 	tmap.lookup = map->lookup;
 	tmap.mc = map->mc;
 	tmap.instance = map->instance;
+	tmap.exp_timeout = map->exp_timeout;
 	tmap.recurse = map->recurse;
 	tmap.depth = map->depth;
+	tmap.stale = map->stale;
 	tmap.argc = 0;
 	tmap.argv = NULL;
 
@@ -585,7 +604,7 @@ int lookup_ghost(struct autofs_point *ap, const char *root)
 		cache_readlock(mc);
 		me = cache_enumerate(mc, NULL);
 		while (me) {
-			if (*me->key == '*')
+			if (!strcmp(me->key, "*"))
 				goto next;
 
 			if (*me->key == '/') {
@@ -673,7 +692,7 @@ static int lookup_name_file_source_instance(struct autofs_point *ap, struct map_
 	char *type, *format;
 
 	if (stat(map->argv[0], &st) == -1) {
-		warn(ap->logopt, "file map not found");
+		debug(ap->logopt, "file map not found");
 		return NSS_STATUS_NOTFOUND;
 	}
 
@@ -752,6 +771,7 @@ static enum nsswitch_status lookup_map_name(struct nss_source *this,
 	tmap.format = map->format;
 	tmap.mc = map->mc;
 	tmap.instance = map->instance;
+	tmap.exp_timeout = map->exp_timeout;
 	tmap.recurse = map->recurse;
 	tmap.depth = map->depth;
 	tmap.argc = 0;
@@ -798,6 +818,10 @@ static void update_negative_cache(struct autofs_point *ap, struct map_source *so
 	struct map_source *map;
 	struct mapent *me;
 
+	/* Don't update negative cache for included maps */ 
+	if (source && source->depth)
+		return;
+
 	/* Have we recorded the lookup fail for negative caching? */
 	me = lookup_source_mapent(ap, name, LKP_DISTINCT);
 	if (me)
@@ -808,7 +832,7 @@ static void update_negative_cache(struct autofs_point *ap, struct map_source *so
 		cache_unlock(me->mc);
 	else {
 		/* Notify only once after fail */
-		error(ap->logopt, "key \"%s\" not found in map.", name);
+		logmsg("key \"%s\" not found in map source(s).", name);
 
 		/* Doesn't exist in any source, just add it somewhere */
 		if (source)
@@ -1014,13 +1038,18 @@ void lookup_prune_one_cache(struct autofs_point *ap, struct mapent_cache *mc, ti
 		char *key = NULL, *next_key = NULL;
 
 		if (me->age >= age) {
+			/*
+			 * Reset time of last fail for valid map entries to
+			 * force entry update and subsequent mount retry.
+			 */
+			me->status = 0;
 			me = cache_enumerate(mc, me);
 			continue;
 		}
 
 		key = strdup(me->key);
 		me = cache_enumerate(mc, me);
-		if (!key || *key == '*') {
+		if (!key || !strcmp(key, "*")) {
 			if (key)
 				free(key);
 			continue;
@@ -1040,6 +1069,14 @@ void lookup_prune_one_cache(struct autofs_point *ap, struct mapent_cache *mc, ti
 		 * cache entry.
 		 */
 		valid = lookup_source_valid_mapent(ap, key, LKP_DISTINCT);
+		if (valid && valid->mc == mc) {
+			 /*
+			  * We've found a map entry that has been removed from
+			  * the current cache so it isn't really valid.
+			  */
+			cache_unlock(valid->mc);
+			valid = NULL;
+		}
 		if (!valid &&
 		    is_mounted(_PATH_MOUNTED, path, MNTS_REAL)) {
 			debug(ap->logopt,
@@ -1111,6 +1148,7 @@ int lookup_prune_cache(struct autofs_point *ap, time_t age)
 		cache_readlock(map->mc);
 		lookup_prune_one_cache(ap, map->mc, age);
 		pthread_cleanup_pop(1);
+		clear_stale_instances(map);
 		map->stale = 0;
 		map = map->next;
 	}
