@@ -400,6 +400,9 @@ static void do_readmap_mount(struct autofs_point *ap, struct mnt_list *mnts,
 		}
 		if (valid) {
 			struct mapent_cache *vmc = valid->mc;
+			struct ioctl_ops *ops = get_ioctl_ops();
+			time_t runfreq;
+
 			cache_unlock(vmc);
 			debug(ap->logopt,
 			     "updating cache entry for valid direct trigger %s",
@@ -412,13 +415,22 @@ static void do_readmap_mount(struct autofs_point *ap, struct mnt_list *mnts,
 			/* Set device and inode number of the new mapent */
 			cache_set_ino_index(vmc, me->key, me->dev, me->ino);
 			cache_unlock(vmc);
+			/* Set timeout and calculate the expire run frequency */
+			ops->timeout(ap->logopt, valid->ioctlfd, map->exp_timeout);
+			if (map->exp_timeout) {
+				runfreq = (map->exp_timeout + CHECK_RATIO - 1) / CHECK_RATIO;
+				if (ap->exp_runfreq)
+					ap->exp_runfreq = min(ap->exp_runfreq, runfreq);
+				else
+					ap->exp_runfreq = runfreq;
+			}
 		} else if (!tree_is_mounted(mnts, me->key, MNTS_REAL))
 			do_umount_autofs_direct(ap, mnts, me);
 		else
 			debug(ap->logopt,
 			      "%s is mounted", me->key);
 	} else
-		do_mount_autofs_direct(ap, mnts, me);
+		do_mount_autofs_direct(ap, mnts, me, map->exp_timeout);
 
 	return;
 }
@@ -466,6 +478,10 @@ static void *do_readmap(void *arg)
 	pthread_cleanup_pop(1);
 
 	if (ap->type == LKP_INDIRECT) {
+		struct ioctl_ops *ops = get_ioctl_ops();
+		time_t timeout = ap->entry->maps->exp_timeout;
+		ap->exp_runfreq = (timeout + CHECK_RATIO - 1) / CHECK_RATIO;
+		ops->timeout(ap->logopt, ap->ioctlfd, timeout);
 		lookup_prune_cache(ap, now);
 		status = lookup_ghost(ap, ap->path);
 	} else {
