@@ -65,9 +65,8 @@ void master_mutex_lock_cleanup(void *arg)
 	return;
 }
 
-int master_add_autofs_point(struct master_mapent *entry, time_t timeout,
-			    unsigned logopt, unsigned nobind, unsigned ghost,
-			    int submount)
+int master_add_autofs_point(struct master_mapent *entry, unsigned logopt,
+			    unsigned nobind, unsigned ghost, int submount)
 {
 	struct autofs_point *ap;
 	int status;
@@ -91,7 +90,6 @@ int master_add_autofs_point(struct master_mapent *entry, time_t timeout,
 	ap->entry = entry;
 	ap->exp_thread = 0;
 	ap->readmap_thread = 0;
-	ap->exp_timeout = timeout;
 	/*
 	 * Program command line option overrides config.
 	 * We can't use 0 negative timeout so use default.
@@ -100,7 +98,7 @@ int master_add_autofs_point(struct master_mapent *entry, time_t timeout,
 		ap->negative_timeout = defaults_get_negative_timeout();
 	else
 		ap->negative_timeout = global_negative_timeout;
-	ap->exp_runfreq = (timeout + CHECK_RATIO - 1) / CHECK_RATIO;
+	ap->exp_runfreq = 0;
 	ap->flags = 0;
 	if (ghost)
 		ap->flags = MOUNT_FLAG_GHOST;
@@ -437,6 +435,7 @@ master_add_source_instance(struct map_source *source, const char *type, const ch
 	new->age = age;
 	new->master_line = 0;
 	new->mc = source->mc;
+	new->exp_timeout = source->exp_timeout;
 	new->stale = 1;
 
 	tmpargv = copy_argv(argc, argv);
@@ -540,38 +539,44 @@ void send_map_update_request(struct autofs_point *ap)
 
 void master_source_writelock(struct master_mapent *entry)
 {
-	int retries = 5; /* 1 second maximum */
+	int status;
+
+	status = pthread_rwlock_wrlock(&entry->source_lock);
+	if (status) {
+		logmsg("master_mapent source write lock failed");
+		fatal(status);
+	}
+	return;
+}
+
+void master_source_readlock(struct master_mapent *entry)
+{
+	int retries = 25;
 	int status;
 
 	while (retries--) {
-		status = pthread_rwlock_wrlock(&entry->source_lock);
-		if (status != EAGAIN)
+		status = pthread_rwlock_rdlock(&entry->source_lock);
+		if (status != EAGAIN && status != EBUSY)
 			break;
 		else {
                 	struct timespec t = { 0, 200000000 };
 	                struct timespec r;
+
+			if (status == EAGAIN)
+				logmsg("master_mapent source too many readers");
+			else
+				logmsg("master_mapent source write lock held");
+
                 	while (nanosleep(&t, &r) == -1 && errno == EINTR)
                         	memcpy(&t, &r, sizeof(struct timespec));
 		}
 	}
 
-	if (status) {
-		logmsg("master_mapent source write lock failed");
-		fatal(status);
-	}
-
-	return;
-}
-
-void master_source_readlock(struct master_mapent *entry)
-{
-	int status;
-
-	status = pthread_rwlock_rdlock(&entry->source_lock);
 	if (status) {
 		logmsg("master_mapent source read lock failed");
 		fatal(status);
 	}
+
 	return;
 }
 
@@ -1283,6 +1288,15 @@ int master_show_mounts(struct master *master)
 		printf("\nMount point: %s\n", ap->path);
 		printf("\nsource(s):\n");
 
+		/*
+		 * Ensure we actually read indirect map entries so we can
+		 * list them. The map reads won't read any indirect map
+		 * entries (other than those in a file map) unless the
+		 * browse option is set.
+		 */
+		if (ap->type == LKP_INDIRECT)
+			ap->flags |= MOUNT_FLAG_GHOST;
+
 		/* Read the map content into the cache */
 		if (lookup_nss_read_map(ap, NULL, now))
 			lookup_prune_cache(ap, now);
