@@ -19,6 +19,8 @@
 #include <sys/stat.h>
 #include <sys/ioctl.h>
 #include <sys/mount.h>
+#include <sys/wait.h>
+#include <ctype.h>
 #include <stdio.h>
 #include <dirent.h>
 #include <sys/vfs.h>
@@ -30,6 +32,8 @@
 #define MAX_OPTIONS_LEN		80
 #define MAX_MNT_NAME_LEN	30
 
+#define EBUFSIZ 1024
+
 const unsigned int t_indirect = AUTOFS_TYPE_INDIRECT;
 const unsigned int t_direct = AUTOFS_TYPE_DIRECT;
 const unsigned int t_offset = AUTOFS_TYPE_OFFSET;
@@ -42,6 +46,35 @@ static const char mnt_name_template[]      = "automount(pid%u)";
 static struct kernel_mod_version kver = {0, 0};
 static const char kver_options_template[]  = "fd=%d,pgrp=%u,minproto=3,maxproto=5";
 
+unsigned int linux_version_code(void)
+{
+	struct utsname my_utsname;
+	unsigned int p, q, r;
+	char *tmp, *save;
+
+	if (uname(&my_utsname))
+		return 0;
+
+	p = q = r = 0;
+
+	tmp = strtok_r(my_utsname.release, ".", &save);
+	if (!tmp)
+		return 0;
+	p = (unsigned int ) atoi(tmp);
+
+	tmp = strtok_r(NULL, ".", &save);
+	if (!tmp)
+		return KERNEL_VERSION(p, 0, 0);
+	q = (unsigned int) atoi(tmp);
+
+	tmp = strtok_r(NULL, ".", &save);
+	if (!tmp)
+		return KERNEL_VERSION(p, q, 0);
+	r = (unsigned int) atoi(tmp);
+
+	return KERNEL_VERSION(p, q, r);
+}
+
 unsigned int query_kproto_ver(void)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
@@ -131,6 +164,142 @@ unsigned int get_kver_minor(void)
 	return kver.minor;
 }
 
+#ifdef HAVE_MOUNT_NFS
+static int extract_version(char *start, struct nfs_mount_vers *vers)
+{
+	char *s_ver = strchr(start, ' ');
+	while (*s_ver && !isdigit(*s_ver)) {
+		s_ver++;
+		if (!*s_ver)
+			return 0;
+		break;
+	}
+	vers->major = atoi(strtok(s_ver, "."));
+	vers->minor = (unsigned int) atoi(strtok(NULL, "."));
+	vers->fix = (unsigned int) atoi(strtok(NULL, "."));
+	return 1;
+}
+
+int check_nfs_mount_version(struct nfs_mount_vers *vers,
+			    struct nfs_mount_vers *check)
+{
+	pid_t f;
+	int ret, status, pipefd[2];
+	char errbuf[EBUFSIZ + 1], *p, *sp;
+	int errp, errn;
+	sigset_t allsigs, tmpsig, oldsig;
+	char *s_ver;
+	int cancel_state;
+
+	if (pipe(pipefd))
+		return -1;
+
+	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel_state);
+
+	sigfillset(&allsigs);
+	pthread_sigmask(SIG_BLOCK, &allsigs, &oldsig);
+
+	f = fork();
+	if (f == 0) {
+		reset_signals();
+		close(pipefd[0]);
+		dup2(pipefd[1], STDOUT_FILENO);
+		dup2(pipefd[1], STDERR_FILENO);
+		close(pipefd[1]);
+
+		execl(PATH_MOUNT_NFS, PATH_MOUNT_NFS, "-V", (char *) NULL);
+		_exit(255);	/* execv() failed */
+	}
+
+	ret = 0;
+
+	tmpsig = oldsig;
+
+	sigaddset(&tmpsig, SIGCHLD);
+	pthread_sigmask(SIG_SETMASK, &tmpsig, NULL);
+
+	close(pipefd[1]);
+
+	if (f < 0) {
+		close(pipefd[0]);
+		pthread_sigmask(SIG_SETMASK, &oldsig, NULL);
+		pthread_setcancelstate(cancel_state, NULL);
+		return -1;
+	}
+
+	errp = 0;
+	do {
+		while (1) {
+			errn = read(pipefd[0], errbuf + errp, EBUFSIZ - errp);
+			if (errn == -1 && errno == EINTR)
+				continue;
+			break;
+		}
+
+		if (errn > 0) {
+			errp += errn;
+
+			sp = errbuf;
+			while (errp && (p = memchr(sp, '\n', errp))) {
+				*p++ = '\0';
+				errp -= (p - sp);
+				sp = p;
+			}
+
+			if (errp && sp != errbuf)
+				memmove(errbuf, sp, errp);
+
+			if (errp >= EBUFSIZ) {
+				/* Line too long, split */
+				errbuf[errp] = '\0';
+				if ((s_ver = strstr(errbuf, "nfs-utils"))) {
+					if (extract_version(s_ver, vers))
+						ret = 1;
+				}
+				errp = 0;
+			}
+
+			if ((s_ver = strstr(errbuf, "nfs-utils"))) {
+				if (extract_version(s_ver, vers))
+					ret = 1;
+			}
+		}
+	} while (errn > 0);
+
+	close(pipefd[0]);
+
+	if (errp > 0) {
+		/* End of file without \n */
+		errbuf[errp] = '\0';
+		if ((s_ver = strstr(errbuf, "nfs-utils"))) {
+			if (extract_version(s_ver, vers))
+				ret = 1;
+		}
+	}
+
+	if (ret) {
+		if ((vers->major < check->major) ||
+		    ((vers->major == check->major) && (vers->minor < check->minor)) ||
+		    ((vers->major == check->major) && (vers->minor == check->minor) &&
+		     (vers->fix < check->fix)))
+			ret = 0;
+	}
+
+	if (waitpid(f, &status, 0) != f) ;
+
+	pthread_sigmask(SIG_SETMASK, &oldsig, NULL);
+	pthread_setcancelstate(cancel_state, NULL);
+
+	return ret;
+}
+#else
+int check_nfs_mount_version(struct nfs_mount_vers *vers,
+			    struct nfs_mount_vers *check)
+{
+	return 0;
+}
+#endif
+
 /*
  * Make common autofs mount options string
  */
@@ -1128,13 +1297,12 @@ const char *mount_type_str(const unsigned int type)
 }
 
 void notify_mount_result(struct autofs_point *ap,
-			 const char *path, const char *type)
+			 const char *path, time_t timeout, const char *type)
 {
-	if (ap->exp_timeout)
+	if (timeout)
 		info(ap->logopt,
 		    "mounted %s on %s with timeout %u, freq %u seconds",
-		    type, path,
-		    (unsigned int) ap->exp_timeout,
+		    type, path, (unsigned int) timeout,
 		    (unsigned int) ap->exp_runfreq);
 	else
 		info(ap->logopt,
@@ -1208,6 +1376,10 @@ static int do_remount_indirect(struct autofs_point *ap, int fd, const char *path
 			int i, j;
 
 			i = j = scandir(buf, &de2, 0, alphasort);
+			if (i < 0) {
+				free(de[n]);
+				continue;
+			}
 			while (i--)
 				free(de2[i]);
 			free(de2);
@@ -1238,16 +1410,14 @@ static int do_remount_indirect(struct autofs_point *ap, int fd, const char *path
 }
 
 static int remount_active_mount(struct autofs_point *ap,
-				struct mapent_cache *mc,
-				const char *path, dev_t devid,
-				const unsigned int type,
-				int *ioctlfd)
+				struct mapent *me, const char *path, dev_t devid,
+				const unsigned int type, int *ioctlfd)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
-	time_t timeout = ap->exp_timeout;
 	const char *str_type = mount_type_str(type);
 	char buf[MAX_ERR_BUF];
 	unsigned int mounted;
+	time_t timeout;
 	struct stat st;
 	int fd;
 
@@ -1257,10 +1427,17 @@ static int remount_active_mount(struct autofs_point *ap,
 	ops->open(ap->logopt, &fd, devid, path);
 	if (fd == -1)
 		return REMOUNT_OPEN_FAIL;
+	else {
+		if (type == t_indirect || type == t_offset)
+			timeout = ap->entry->maps->exp_timeout;
+		else
+			timeout = me->source->exp_timeout;
+	}
 
 	/* Re-reading the map, set timeout and return */
 	if (ap->state == ST_READMAP) {
-		ops->timeout(ap->logopt, fd, &timeout);
+		debug(ap->logopt, "already mounted, update timeout");
+		ops->timeout(ap->logopt, fd, timeout);
 		ops->close(ap->logopt, fd);
 		return REMOUNT_READ_MAP;
 	}
@@ -1282,7 +1459,7 @@ static int remount_active_mount(struct autofs_point *ap,
 		ops->close(ap->logopt, fd);
 		return REMOUNT_OPEN_FAIL;
 	}
-	ops->timeout(ap->logopt, fd, &timeout);
+	ops->timeout(ap->logopt, fd, timeout);
 	if (fstat(fd, &st) == -1) {
 		error(ap->logopt,
 		      "failed to stat %s mount %s", str_type, path);
@@ -1290,11 +1467,11 @@ static int remount_active_mount(struct autofs_point *ap,
 		ops->close(ap->logopt, fd);
 		return REMOUNT_STAT_FAIL;
 	}
-	if (mc)
-		cache_set_ino_index(mc, path, st.st_dev, st.st_ino);
+	if (type != t_indirect)
+		cache_set_ino_index(me->mc, path, st.st_dev, st.st_ino);
 	else
 		ap->dev = st.st_dev;
-	notify_mount_result(ap, path, str_type);
+	notify_mount_result(ap, path, timeout, str_type);
 
 	*ioctlfd = fd;
 
@@ -1337,24 +1514,20 @@ static int remount_active_mount(struct autofs_point *ap,
 int try_remount(struct autofs_point *ap, struct mapent *me, unsigned int type)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
-	struct mapent_cache *mc;
 	const char *path;
 	int ret, fd;
 	dev_t devid;
 
-	if (type == t_indirect) {
-		mc = NULL;
+	if (type == t_indirect)
 		path = ap->path;
-	} else {
-		mc = me->mc;
+	else
 		path = me->key;
-	}
 
 	ret = ops->mount_device(ap->logopt, path, type, &devid);
 	if (ret == -1 || ret == 0)
 		return -1;
 
-	ret = remount_active_mount(ap, mc, path, devid, type, &fd);
+	ret = remount_active_mount(ap, me, path, devid, type, &fd);
 
 	/*
 	 * The directory must exist since we found a device
@@ -1378,7 +1551,9 @@ int try_remount(struct autofs_point *ap, struct mapent *me, unsigned int type)
 	 * record that in the mount point struct. Otherwise we're
 	 * re-reading the map.
 	*/
-	if (ret == REMOUNT_SUCCESS || ret == REMOUNT_READ_MAP) {
+	if (ret == REMOUNT_READ_MAP)
+		return 1;
+	else if (ret == REMOUNT_SUCCESS) {
 		if (fd != -1) {
 			if (type == t_indirect)
 				ap->ioctlfd = fd;
@@ -1433,6 +1608,33 @@ int umount_ent(struct autofs_point *ap, const char *path)
 	return rv;
 }
 
+static int do_mount_autofs_offset(struct autofs_point *ap,
+				  struct mapent *oe, const char *root,
+				  char *offset)
+
+{
+	int mounted = 0;
+	int ret;
+
+	debug(ap->logopt, "mount offset %s at %s", oe->key, root);
+
+	ret = mount_autofs_offset(ap, oe, root, offset);
+	if (ret >= MOUNT_OFFSET_OK)
+		mounted++;
+	else {
+		if (ret != MOUNT_OFFSET_IGNORE)
+			warn(ap->logopt, "failed to mount offset");
+		else {
+			debug(ap->logopt, "ignoring \"nohide\" trigger %s",
+			      oe->key);
+			free(oe->mapent);
+			oe->mapent = NULL;
+		}
+	}
+
+	return mounted;
+}
+
 int mount_multi_triggers(struct autofs_point *ap, struct mapent *me,
 			 const char *root, unsigned int start, const char *base)
 {
@@ -1441,8 +1643,7 @@ int mount_multi_triggers(struct autofs_point *ap, struct mapent *me,
 	struct mapent *oe;
 	struct list_head *pos = NULL;
 	unsigned int fs_path_len;
-	unsigned int mounted;
-	int ret;
+	int mounted;
 
 	fs_path_len = start + strlen(base);
 	if (fs_path_len > PATH_MAX)
@@ -1462,20 +1663,20 @@ int mount_multi_triggers(struct autofs_point *ap, struct mapent *me,
 		if (!oe || !oe->mapent)
 			goto cont;
 
-		debug(ap->logopt, "mount offset %s at %s", oe->key, root);
+		mounted += do_mount_autofs_offset(ap, oe, root, offset);
 
-		ret = mount_autofs_offset(ap, oe, root, offset);
-		if (ret >= MOUNT_OFFSET_OK)
-			mounted++;
-		else {
-			if (ret != MOUNT_OFFSET_IGNORE)
-				warn(ap->logopt, "failed to mount offset");
-			else {
-				debug(ap->logopt,
-				      "ignoring \"nohide\" trigger %s",
-				      oe->key);
-				free(oe->mapent);
-				oe->mapent = NULL;
+		/*
+		 * If re-constructing a multi-mount it's necessary to walk
+		 * into nested mounts, unlike the usual "mount only what's
+		 * needed as you go" behavior.
+		 */
+		if (ap->state == ST_READMAP && ap->flags & MOUNT_FLAG_REMOUNT) {
+			if (oe->ioctlfd != -1 ||
+			    is_mounted(_PROC_MOUNTS, oe->key, MNTS_REAL)) {
+				char oe_root[PATH_MAX + 1];
+				strcpy(oe_root, root);
+				strcat(oe_root, offset); 
+				mount_multi_triggers(ap, oe, oe_root, strlen(oe_root), base);
 			}
 		}
 cont:
@@ -1486,6 +1687,41 @@ cont:
 	return mounted;
 }
 
+static int rmdir_path_offset(struct autofs_point *ap, struct mapent *oe)
+{
+	char *dir, *path;
+	unsigned int split;
+	int ret;
+
+	if (ap->type == LKP_DIRECT)
+		return rmdir_path(ap, oe->key, oe->multi->dev);
+
+	dir = strdup(oe->key);
+
+	if (ap->flags & MOUNT_FLAG_GHOST)
+		split = strlen(ap->path) + strlen(oe->multi->key) + 1;
+	else
+		split = strlen(ap->path);
+
+	dir[split] = '\0';
+	path = &dir[split + 1];
+
+	if (chdir(dir) == -1) {
+		error(ap->logopt, "failed to chdir to %s", dir);
+		free(dir);
+		return -1;
+	}
+
+	ret = rmdir_path(ap, path, ap->dev);
+
+	free(dir);
+
+	if (chdir("/") == -1)
+		error(ap->logopt, "failed to chdir to /");
+
+	return ret;
+}
+
 int umount_multi_triggers(struct autofs_point *ap, struct mapent *me, char *root, const char *base)
 {
 	char path[PATH_MAX + 1];
@@ -1509,7 +1745,6 @@ int umount_multi_triggers(struct autofs_point *ap, struct mapent *me, char *root
 	pos = NULL;
 	offset = path;
 
-	/* Make sure "none" of the offsets have an active mount. */
 	while ((offset = cache_get_offset(mm_base, offset, start, mm_root, &pos))) {
 		char *oe_base;
 
@@ -1528,28 +1763,34 @@ int umount_multi_triggers(struct autofs_point *ap, struct mapent *me, char *root
 		if (oe->ioctlfd != -1 ||
 		    is_mounted(_PROC_MOUNTS, oe->key, MNTS_REAL)) {
 			left++;
-			break;
-		}
-	}
-
-	if (left)
-		return left;
-
-	pos = NULL;
-	offset = path;
-
-	/* Make sure "none" of the offsets have an active mount. */
-	while ((offset = cache_get_offset(mm_base, offset, start, mm_root, &pos))) {
-		oe = cache_lookup_offset(mm_base, offset, start, &me->multi_list);
-		/* root offset is a special case */
-		if (!oe || !oe->mapent || (strlen(oe->key) - start) == 1)
 			continue;
+		}
 
 		debug(ap->logopt, "umount offset %s", oe->key);
 
 		if (umount_autofs_offset(ap, oe)) {
 			warn(ap->logopt, "failed to umount offset");
 			left++;
+		} else {
+			struct stat st;
+			int ret;
+
+			if (!(oe->flags & MOUNT_FLAG_DIR_CREATED))
+				continue;
+
+			/*
+			 * An error due to partial directory removal is
+			 * ok so only try and remount the offset if the
+			 * actual mount point still exists.
+			 */
+			ret = rmdir_path_offset(ap, oe);
+			if (ret == -1 && !stat(oe->key, &st)) {
+				ret = do_mount_autofs_offset(ap, oe, root, offset);
+				if (ret)
+					left++;
+				/* But we did origianlly create this */
+				oe->flags |= MOUNT_FLAG_DIR_CREATED;
+			}
 		}
 	}
 
@@ -1582,3 +1823,162 @@ int umount_multi_triggers(struct autofs_point *ap, struct mapent *me, char *root
 	return left;
 }
 
+int clean_stale_multi_triggers(struct autofs_point *ap,
+			       struct mapent *me, char *top, const char *base)
+{
+	char *root;
+	char mm_top[PATH_MAX + 1];
+	char path[PATH_MAX + 1];
+	char buf[MAX_ERR_BUF];
+	char *offset;
+	struct mapent *oe;
+	struct list_head *mm_root, *pos;
+	const char o_root[] = "/";
+	const char *mm_base;
+	int left, start;
+	time_t age;
+
+	if (top)
+		root = top;
+	else {
+		if (!strchr(me->multi->key, '/'))
+			/* Indirect multi-mount root */
+			/* sprintf okay - if it's mounted, it's
+			 * PATH_MAX or less bytes */
+			sprintf(mm_top, "%s/%s", ap->path, me->multi->key);
+		else
+			strcpy(mm_top, me->multi->key);
+		root = mm_top;
+	}
+
+	left = 0;
+	start = strlen(root);
+
+	mm_root = &me->multi->multi_list;
+
+	if (!base)
+		mm_base = o_root;
+	else
+		mm_base = base;
+
+	pos = NULL;
+	offset = path;
+	age = me->multi->age;
+
+	while ((offset = cache_get_offset(mm_base, offset, start, mm_root, &pos))) {
+		char *oe_base;
+		char *key;
+		int ret;
+
+		oe = cache_lookup_offset(mm_base, offset, start, &me->multi_list);
+		/* root offset is a special case */
+		if (!oe || !oe->mapent || (strlen(oe->key) - start) == 1)
+			continue;
+
+		/* Check for and umount stale subtree offsets */
+		oe_base = oe->key + strlen(root);
+		ret = clean_stale_multi_triggers(ap, oe, root, oe_base);
+		left =+ ret;
+		if (ret)
+			continue;
+
+		if (oe->age == age)
+			continue;
+
+		/*
+		 * If an offset that has an active mount has been removed
+		 * from the multi-mount we don't want to attempt to trigger
+		 * mounts for it. Obviously this is because it has been
+		 * removed, but less obvious is the potential strange
+		 * behaviour that can result if we do try and mount it
+		 * again after it's been expired. For example, if an NFS
+		 * file system is no longer exported and is later umounted
+		 * it can be mounted again without any error message but
+		 * shows as an empty directory. That's going to confuse
+		 * people for sure.
+		 *
+		 * If the mount cannot be umounted (the process is now
+		 * using a stale mount) the offset needs to be invalidated
+		 * so no further mounts will be attempted but the offset
+		 * cache entry must remain so expires can continue to
+		 * attempt to umount it. If the mount can be umounted and
+		 * the offset is removed, at least for NFS we will get
+		 * ESTALE errors when attempting list the directory.
+		 */
+		if (oe->ioctlfd != -1 ||
+		    is_mounted(_PROC_MOUNTS, oe->key, MNTS_REAL)) {
+			if (umount_ent(ap, oe->key)) {
+				debug(ap->logopt,
+				      "offset %s has active mount, invalidate",
+				      oe->key);
+				if (oe->mapent) {
+					free(oe->mapent);
+					oe->mapent = NULL;
+				}
+				left++;
+				continue;
+			}
+		}
+
+		key = strdup(oe->key);
+		if (!key) {
+	                char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+        	        error(ap->logopt, "malloc: %s", estr);
+			left++;
+			continue;
+		}
+
+		debug(ap->logopt, "umount offset %s", oe->key);
+
+		if (umount_autofs_offset(ap, oe)) {
+			warn(ap->logopt, "failed to umount offset %s", key);
+			left++;
+		} else {
+			struct stat st;
+
+			/* Mount point not ours to delete ? */
+			if (!(oe->flags & MOUNT_FLAG_DIR_CREATED)) {
+				debug(ap->logopt, "delete offset key %s", key);
+				if (cache_delete_offset(oe->mc, key) == CHE_FAIL)
+					error(ap->logopt,
+					     "failed to delete offset key %s", key);
+				free(key);
+				continue;
+			}
+
+			/*
+			 * An error due to partial directory removal is
+			 * ok so only try and remount the offset if the
+			 * actual mount point still exists.
+			 */
+			ret = rmdir_path_offset(ap, oe);
+			if (ret == -1 && !stat(oe->key, &st)) {
+				ret = do_mount_autofs_offset(ap, oe, root, offset);
+				if (ret) {
+					left++;
+					/* But we did origianlly create this */
+					oe->flags |= MOUNT_FLAG_DIR_CREATED;
+					free(key);
+					continue;
+				}
+				/*
+				 * Fall through if the trigger can't be mounted
+				 * again, since there is no offset there can't
+				 * be any mount requests so remove the map
+				 * entry from the cache. There's now a dead
+				 * offset mount, but what else can we do ....
+				 */
+			}
+
+			debug(ap->logopt, "delete offset key %s", key);
+
+			if (cache_delete_offset(oe->mc, key) == CHE_FAIL)
+				error(ap->logopt,
+				     "failed to delete offset key %s", key);
+		}
+		free(key);
+	}
+
+	return left;
+}
+
