@@ -280,8 +280,7 @@ static int get_query_dn(unsigned logopt, LDAP *ldap, struct lookup_context *ctxt
 {
 	char buf[MAX_ERR_BUF];
 	char *query, *dn, *qdn;
-	LDAPMessage *result, *e;
-	struct ldap_searchdn *sdns = NULL;
+	LDAPMessage *result = NULL, *e;
 	char *attrs[2];
 	struct berval **value;
 	int scope;
@@ -330,15 +329,6 @@ static int get_query_dn(unsigned logopt, LDAP *ldap, struct lookup_context *ctxt
 		scope = LDAP_SCOPE_SUBTREE;
 	}
 
-	if (!ctxt->base) {
-		sdns = defaults_get_searchdns();
-		if (sdns) {
-			if (ctxt->sdns)
-				defaults_free_searchdns(ctxt->sdns);
-			ctxt->sdns = sdns;
-		}
-	}
-
 	dn = NULL;
 	if (!ctxt->sdns) {
 		rv = ldap_search_s(ldap, ctxt->base,
@@ -347,6 +337,8 @@ static int get_query_dn(unsigned logopt, LDAP *ldap, struct lookup_context *ctxt
 			error(logopt,
 			      MODPREFIX "query failed for %s: %s",
 			      query, ldap_err2string(rv));
+			if (result)
+				ldap_msgfree(result);
 			free(query);
 			return 0;
 		}
@@ -519,7 +511,7 @@ static int do_bind(unsigned logopt, LDAP *ldap, const char *uri, struct lookup_c
 	debug(logopt, MODPREFIX "auth_required: %d, sasl_mech %s",
 	      ctxt->auth_required, ctxt->sasl_mech);
 
-	if (ctxt->auth_required & (LDAP_AUTH_REQUIRED|LDAP_AUTH_AUTODETECT)) {
+	if (ctxt->auth_required & LDAP_NEED_AUTH) {
 		rv = autofs_sasl_bind(logopt, ldap, ctxt);
 		debug(logopt, MODPREFIX "autofs_sasl_bind returned %d", rv);
 	} else {
@@ -734,13 +726,12 @@ static LDAP *find_server(unsigned logopt, struct lookup_context *ctxt)
 static LDAP *do_reconnect(unsigned logopt, struct lookup_context *ctxt)
 {
 	LDAP *ldap = NULL;
-	char *uri;
 
 	if (ctxt->server || !ctxt->uris) {
 		ldap = do_connect(logopt, ctxt->server, ctxt);
 #ifdef WITH_SASL
 		/* Dispose of the sasl authentication connection and try again. */
-		if (!ldap) {
+		if (!ldap && ctxt->auth_required & LDAP_NEED_AUTH) {
 			autofs_sasl_dispose(ctxt);
 			ldap = connect_to_server(logopt, ctxt->server, ctxt);
 		}
@@ -776,9 +767,9 @@ static LDAP *do_reconnect(unsigned logopt, struct lookup_context *ctxt)
 	 * Dispose of the sasl authentication connection and try the
 	 * current server again before trying other servers in the list.
 	 */
-	if (!ldap) {
+	if (!ldap && ctxt->auth_required & LDAP_NEED_AUTH) {
 		autofs_sasl_dispose(ctxt);
-		ldap = connect_to_server(logopt, uri, ctxt);
+		ldap = connect_to_server(logopt, ctxt->uri->uri, ctxt);
 	}
 #endif
 	if (ldap)
@@ -1466,6 +1457,9 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 		return 1;
 	}
 
+	if (!ctxt->base)
+		ctxt->sdns = defaults_get_searchdns();
+
 	ctxt->timeout = defaults_get_ldap_timeout();
 	ctxt->network_timeout = defaults_get_ldap_network_timeout();
 
@@ -1527,7 +1521,7 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 	char buf[MAX_ERR_BUF];
 	char parse_buf[PARSE_MAX_BUF];
 	char *query;
-	LDAPMessage *result, *e;
+	LDAPMessage *result = NULL, *e;
 	char *class, *info, *entry;
 	char **keyValue = NULL;
 	char **values = NULL;
@@ -1573,6 +1567,8 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 		error(logopt, MODPREFIX "query failed for %s: %s",
 		      query, ldap_err2string(rv));
 		unbind_ldap_connection(logging, ldap, ctxt);
+		if (result)
+			ldap_msgfree(result);
 		free(query);
 		return NSS_STATUS_NOTFOUND;
 	}
@@ -2041,7 +2037,8 @@ do_paged:
 	rv = ldap_parse_page_control(sp->ldap,
 				     returnedControls, &sp->totalCount,
 				     &sp->cookie);
-	if (sp->cookie && sp->cookie->bv_val && strlen(sp->cookie->bv_val))
+	if (sp->cookie && sp->cookie->bv_val &&
+	    (strlen(sp->cookie->bv_val) || sp->cookie->bv_len))
 		sp->morePages = TRUE;
 	else
 		sp->morePages = FALSE;
@@ -2329,8 +2326,10 @@ static int read_one_map(struct autofs_point *ap,
 	 * reading the map. We always need to read the whole map for
 	 * direct mounts in order to mount the triggers.
 	 */
-	if (!(ap->flags & MOUNT_FLAG_GHOST) && ap->type != LKP_DIRECT)
+	if (!(ap->flags & MOUNT_FLAG_GHOST) && ap->type != LKP_DIRECT) {
+		debug(ap->logopt, "map read not needed, so not done");
 		return NSS_STATUS_SUCCESS;
+	}
 
 	sp.ap = ap;
 	sp.age = age;
@@ -2380,8 +2379,14 @@ static int read_one_map(struct autofs_point *ap,
 
 		if (rv == LDAP_ADMINLIMIT_EXCEEDED ||
 		    rv == LDAP_SIZELIMIT_EXCEEDED) {
-			if (sp.result)
+			if (sp.result) {
 				ldap_msgfree(sp.result);
+				sp.result = NULL;
+			}
+			if (sp.cookie) {
+				ber_bvfree(sp.cookie);
+				sp.cookie = NULL;
+			}
 			sp.pageSize = sp.pageSize / 2;
 			if (sp.pageSize < 5) {
 				debug(ap->logopt, MODPREFIX
@@ -2397,6 +2402,10 @@ static int read_one_map(struct autofs_point *ap,
 		if (rv != LDAP_SUCCESS || !sp.result) {
 			unbind_ldap_connection(ap->logopt, sp.ldap, ctxt);
 			*result_ldap = rv;
+			if (sp.result)
+				ldap_msgfree(sp.result);
+			if (sp.cookie)
+				ber_bvfree(sp.cookie);
 			free(sp.query);
 			return NSS_STATUS_UNAVAIL;
 		}
@@ -2406,10 +2415,13 @@ static int read_one_map(struct autofs_point *ap,
 			ldap_msgfree(sp.result);
 			unbind_ldap_connection(ap->logopt, sp.ldap, ctxt);
 			*result_ldap = rv;
+			if (sp.cookie)
+				ber_bvfree(sp.cookie);
 			free(sp.query);
 			return NSS_STATUS_NOTFOUND;
 		}
 		ldap_msgfree(sp.result);
+		sp.result = NULL;
 	} while (sp.morePages == TRUE);
 
 	debug(ap->logopt, MODPREFIX "done updating map");
@@ -2417,6 +2429,8 @@ static int read_one_map(struct autofs_point *ap,
 	unbind_ldap_connection(ap->logopt, sp.ldap, ctxt);
 
 	source->age = age;
+	if (sp.cookie)
+		ber_bvfree(sp.cookie);
 	free(sp.query);
 
 	return NSS_STATUS_SUCCESS;
@@ -2455,7 +2469,7 @@ static int lookup_one(struct autofs_point *ap,
 	char buf[MAX_ERR_BUF];
 	time_t age = time(NULL);
 	char *query;
-	LDAPMessage *result, *e;
+	LDAPMessage *result = NULL, *e;
 	char *class, *info, *entry;
 	char *enc_key1, *enc_key2;
 	int enc_len1 = 0, enc_len2 = 0;
@@ -2570,6 +2584,8 @@ static int lookup_one(struct autofs_point *ap,
 	if ((rv != LDAP_SUCCESS) || !result) {
 		crit(ap->logopt, MODPREFIX "query failed for %s", query);
 		unbind_ldap_connection(ap->logopt, ldap, ctxt);
+		if (result)
+			ldap_msgfree(result);
 		free(query);
 		return CHE_FAIL;
 	}
